#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <poll.h>
#include <netdb.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include <pthread.h>

char *wstr = NULL;
char *rstr = NULL;

int verbose = 0;

int timeout = 1000;
int nr_crlf = 1;
int pr_sym  = 0;

int first_port = 0;
int last_port  = 0;

unsigned long  first_ip = 0x00000000;
unsigned long  last_ip  = 0xFFFFFFFF;

pthread_mutex_t cur_lock;
unsigned long  cur_ip;
int cur_port;

pthread_mutex_t print_lock;

void
print_host(FILE *fp,
	   struct in_addr in,
	   int port)
{
    struct hostent *hep = NULL;

    
    if (pr_sym)
    {
	hep = gethostbyaddr((const char *) &in, sizeof(in), AF_INET);
	fprintf(fp, "%-15s : %-40s : %d", inet_ntoa(in), hep ? hep->h_name : "(unknown)", port);
    }
    else
	fprintf(fp, "%-15s : %d", inet_ntoa(in), port);
}


int
t_write(int fd,
	char *buf,
	int len)
{
    int tw, wl, code;
    struct pollfd pfd;


    tw = len;
    while (tw > 0)
    {
	pfd.fd = fd;
	pfd.events = POLLOUT;
	pfd.revents = 0;
	
	while ((code = poll(&pfd, 1, timeout)) < 0 && errno == EINTR)
	    errno = 0;

	if (code == 0)
	{
	    code = -1;
	    errno = ETIMEDOUT;
	}
	
	while ((wl = write(fd, buf, tw)) < 0 && errno == EINTR)
	    ;
	
	if (wl < 0)
	    return wl;

	tw -= wl;
	buf += wl;
    }

    return len;
}


int
t_read(int fd,
       char *buf,
       int size)
{
    int len, code;
    struct pollfd pfd;


    pfd.fd = fd;
    pfd.events = POLLIN;
    pfd.revents = 0;
    
    while ((code = poll(&pfd, 1, timeout)) < 0 && errno == EINTR)
	errno = 0;
    
    if (code == 0)
    {
	errno = ETIMEDOUT;
	return -1;
    }
    
    while ((len = read(fd, buf, size)) < 0 && errno == EINTR)
	;
    
    return len;
}

int
is_text(char *str)
{
    while (*str && (isprint(*str) || *str == '\t' || *str == '\n' || *str == '\r'))
	++str;

    return *str == 0;
}


int
print_output(char *str)
{
    if (str == NULL)
	return -1;

    if (is_text(str))
    {
	printf("TXT : ");
    
	while (*str)
	{
	    if (isprint(*str))
		putchar(*str);
	    else
		switch (*str)
		{
		  case '\n':
		    printf("\\n");
		    break;
		    
		  case '\r':
		    printf("\\r");
		    break;
		    
		  case '\t':
		    printf("\\t");
		    break;
		    
		  default:
		    printf("\\x%02x", * (unsigned char *) str);
		}
	    
	    ++str;
	}
    }
    else
    {
	printf("HEX :");
	while (*str)
	{
	    printf(" %02x", * (unsigned char *) str);
	    ++str;
	}
    }

    return 0;
}


int
check_port(unsigned long addr,
	   int port)
{
    int fd, code, len, n_crlf;
    struct sockaddr_in sin;
    char c, buf[512];
    static int one = 1;
    struct pollfd pfd;
    
    
    len = 0;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0)
	return -1;

    memset(&sin, 0, sizeof(sin));

    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);

    sin.sin_addr.s_addr = addr;

    code = fcntl(fd, F_GETFL, 0);
    if (code < 0)
    {
	close(fd);
	return -1;
    }

#ifdef FNDELAY
    code = fcntl(fd, F_SETFL, code|FNDELAY);
#else
    code = fcntl(fd, F_SETFL, code|O_NONBLOCK);
#endif
    if (code < 0)
    {
	close(fd);
	return -1;
    }
    
    while ((code = connect(fd,
			   (struct sockaddr *) &sin, sizeof(sin))) < 0 &&
	   errno == EINTR)
	errno = 0;
    
    if (code < 0 && errno == EINPROGRESS)
    {
	pfd.fd = fd;
	pfd.events = POLLOUT;
	pfd.revents = 0;
	
	while ((code = poll(&pfd, 1, timeout)) < 0 && errno == EINTR)
	    errno = 0;

	if (code == 0)
	{
	    code = -1;
	    errno = ETIMEDOUT;
	}
    }
    
    if (code < 0)
    {
#if 0
	if (errno == ECONNREFUSED || errno == ETIMEDOUT)
	{
	    close(fd);
	    return 0;
	}
#endif

	if (verbose)
	{
	    pthread_mutex_lock(&print_lock);
	    
	    print_host(stderr, sin.sin_addr, port);
	    fprintf(stderr, " : connect() failed: %s\n", strerror(errno));
	    
	    pthread_mutex_unlock(&print_lock);
	}
	
	close(fd);
	return 0;
    }

    if (wstr)
    {
	code = t_write(fd, wstr, strlen(wstr));
	if (code < 0)
	{
	    close(fd);
	    return 0;
	}
	
	code = t_write(fd, "\r\n\r\n", 4);
	if (code < 0)
	{
	    close(fd);
	    return 0;
	}
    }

    shutdown(fd, 1);

    n_crlf = 0;
    
    while ((code = t_read(fd, &c, 1)) != 0 && len < sizeof(buf)-1)
    {
	if (code < 0)
	{
	    if (errno == EINTR)
		continue;

#if 0
	    if (errno == ECONNREFUSED || errno == ETIMEDOUT ||
		errno == ECONNRESET)
	    {
		close(fd);
		return 0;
	    }
#endif
	    
	    if (verbose)
	    {
		pthread_mutex_lock(&print_lock);
		
		print_host(stderr, sin.sin_addr, port);
		fprintf(stderr, ": read() failed: %s\n", strerror(errno));
		
		pthread_mutex_unlock(&print_lock);
	    }
	    
	    close(fd);
	    return -1;
	}
	
	if (c == '\r' || c == '\n')
	{
	    ++n_crlf;

	    if (n_crlf >= nr_crlf)
		break;
	    
	}
	
	buf[len++] = c;
    }

    buf[len] = '\0';
    
    if (rstr)
    {
	char *lp, *cp;
	char *tokp;
	
	
	lp = strtok_r(buf, "\r\n", &tokp);
	while (lp)
	{
	    if (verbose)
		fprintf(stderr, "[checking: %s]\n", lp);
	    
	    if (strncasecmp(lp, rstr, strlen(rstr)) == 0)
	    {
		pthread_mutex_lock(&print_lock);
		
	        print_host(stdout, sin.sin_addr, port);
		printf(" : ");
		print_output(lp);
		putchar('\n');

		pthread_mutex_unlock(&print_lock);

		break;
	    }
	    
	    lp = strtok_r(NULL, "\r\n", &tokp);
	}
    }
    else
    {
	char *lp, *tokp;

	
	lp = strtok_r(buf, "\r\n", &tokp);

	if (lp)
	{
	    pthread_mutex_lock(&print_lock);
	    
	    print_host(stdout, sin.sin_addr, port);
	    printf(" : ");
	    print_output(lp);
	    putchar('\n');
	    
	    pthread_mutex_unlock(&print_lock);
	}
    }
    
    close(fd);
    return 1;
}


void *
worker(void *arg)
{
    unsigned long addr;
    int port;
    
    
  Again:
    pthread_mutex_lock(&cur_lock);
    if (cur_ip > last_ip)
    {
	++cur_port;
	
	if (cur_port > last_port)
	{
	    pthread_mutex_unlock(&cur_lock);
	    return NULL;
	}

	cur_ip = first_ip;
    }

    port = cur_port;
    addr = cur_ip++;
    
    pthread_mutex_unlock(&cur_lock);

    check_port(htonl(addr), port);
    
    goto Again;
}


char *argv0 = "pnscan";


void
usage(FILE *out)
{
    fprintf(out, "Usage: %s [<options>] <CIDR | host-range> <port-range>\n", argv0);
    fprintf(out, "Options:\n");
    fprintf(out, "\t-h                Display this information.\n");
    fprintf(out, "\t-m                Expect a multiline response.\n");
    fprintf(out, "\t-s                Lookup and print hostnames.\n");
    fprintf(out, "\t-v                Be verbose.\n");
    fprintf(out, "\t-w<string>        Request string to send.\n");
    fprintf(out, "\t-r<string>        Response string to look for.\n");
    fprintf(out, "\t-t<msecs>         Connect/Write/Read timeout.\n");
    fprintf(out, "\t-n<workers>       Concurrent worker threads.\n");
}
    

int
get_network(char *str,
	    unsigned long *np)
{
    struct netent *nep;
    struct in_addr ia;
    

    nep = getnetbyname(str);
    if (nep)
    {
	ia = inet_makeaddr(nep->n_net, 0);
	*np = ntohl(ia.s_addr);
	return 1;
    }

    return inet_pton(AF_INET, str, np);
}


int
get_host(char *str,
	 unsigned long *ip)
{
    struct hostent *hep;
    unsigned long tip;


    hep = gethostbyname(str);
    if (hep && hep->h_addr_list &&hep->h_addr_list[0])
    {
	tip = * (unsigned long *) (hep->h_addr_list[0]);
	*ip = ntohl(tip);
	return 1;
    }

    return inet_pton(AF_INET, str, ip);
}


int
get_service(char *str,
	    int *pp)
{
    struct servent *sep;


    sep = getservbyname(str, "tcp");
    if (sep)
    {
	*pp = ntohs(sep->s_port);
	return 1;
    }

    if (sscanf(str, "%u", pp) != 1)
	return -1;

    if (*pp < 1 || *pp > 65535)
	return 0;

    return 1;
}

int
get_ip_range(char *str,
	     unsigned long *first_ip,
	     unsigned long *last_ip)
{
    char first[256], last[256];
    int len;
    unsigned long ip;
    unsigned long mask;
    

    if (sscanf(str, "%255[^/ ] / %u", first, &len) == 2)
    {
	/* CIDR */

	if (get_network(first, &ip) != 1 || len < 0 || len > 32)
	    return -1;

	*first_ip = ip+1;
	
	len = 32-len;
	while (len-- > 0)
	    mask = ((mask << 1)|1);

	*last_ip = (ip|mask)-1;

	return 2;
    }

    switch (sscanf(str, "%255[^: ] : %255s", first, last))
    {
      case 1:
	if (get_host(first, first_ip) != 1)
	    return -1;

	*last_ip = *first_ip;
	return 1;

      case 2:
	if (get_host(first, first_ip) != 1)
	    return -1;

	if (get_host(last, last_ip) != 1)
	    return -1;

	return 2;
    }
    
    return -1;
}


int
get_port_range(char *str,
	       int *first_port,
	       int *last_port)
{
    char first[256], last[256];


    switch (sscanf(str, "%255[^: ] : %255s", first, last))
    {
      case 1:
	if (get_service(first, first_port) != 1)
	    return -1;

	*last_port = *first_port;
	return 1;

      case 2:
	if (get_service(first, first_port) != 1)
	    return -1;

	if (get_service(last, last_port) != 1)
	    return -1;

	return 2;
    }
    
    return -1;
}


int
main(int argc,
     char *argv[])
{
    int i, nworkers, len;
    pthread_t tid;
    struct rlimit rlb;
    

    argv0 = argv[0];


    first_port = 0;
    last_port = 0;
    
    getrlimit(RLIMIT_NOFILE, &rlb);
    rlb.rlim_cur = rlb.rlim_max;
    setrlimit(RLIMIT_NOFILE, &rlb);

    signal(SIGPIPE, SIG_IGN);
    
    nworkers = rlb.rlim_cur - 8;

    if (nworkers > 1024)
	nworkers = 1024;

    pthread_mutex_init(&cur_lock, NULL);

    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	  case 'h':
	    usage(stdout);
	    exit(0);
    
	  case 'w':
	    wstr = argv[i]+2;
	    break;

	  case 's':
	    ++pr_sym;
	    break;
	    
	  case 'r':
	    rstr = argv[i]+2;
	    break;

	  case 'm':
	    nr_crlf = 50;
	    break;
	    
	  case 't':
	    timeout = atoi(argv[i]+2);
	    break;

	  case 'n':
	    nworkers = atoi(argv[i]+2);
	    break;
	    
	  case 'v':
	    ++verbose;
	    break;
	    
	  default:
	    fprintf(stderr, "%s: unknown command line switch: %s\n",
		    argv[0], argv[i]);
	    exit(1);
	}

    if (get_ip_range(argv[i], &first_ip, &last_ip) < 1)
    {
	fprintf(stderr, "%s: Invalid IP address range: %s\n",
		argv[0], argv[i]);
	exit(1);
    }

    if (get_port_range(argv[i+1], &first_port, &last_port) < 1)
    {
	fprintf(stderr, "%s: Invalid Port range: %s\n",
		argv[0], argv[i+1]);
	exit(1);
    }
	
    cur_ip = first_ip;
    cur_port = first_port;
    
    len = (last_ip - first_ip + 1)*(last_port - first_port + 1);

    if (len < nworkers)
	nworkers = len;
    
    if (verbose)
	fprintf(stderr, "Using %d worker threads\n", nworkers);

    for (i = 0; i < nworkers; ++i)
	pthread_create(&tid, NULL, worker, NULL);

    pthread_exit(NULL);
}
